#!/usr/bin/env python
import numpy as np
from copy import deepcopy

def calculate_cell_area(self):

    # Create bounds if they do not exist
    if self._lat_bnds is None or self._lon_bnds is None:
        self.create_spatial_bounds()

    # Get spatial number of vertices
    spatial_nv = self._lat_bnds.shape[-1]

    # Calculate area
    area = np.empty(shape=(len(self.lon['data']), len(self.lat['data'])))
    if spatial_nv <= 4:
        for lon_ind in range(0, len(self.lon['data'])):
            for lat_ind in range(0, len(self.lat['data'])):
                grid_corner_lon = np.array(self._lon['data'], dtype=np.float32).flatten()
                grid_corner_lat = np.array(self._lat['data'], dtype=np.float32).flatten()
                area[lon_ind, lat_ind] = self.mod_huiliers_area(spatial_nv, 
                                                                grid_corner_lon+lon_ind*spatial_nv, 
                                                                grid_corner_lat+lon_ind*spatial_nv)

    return area

def mod_huiliers_area(self, spatial_nv, cell_corner_lon, cell_corner_lat):
    
    sum = 0

    if spatial_nv >= 3:
        
        xyz1 = self.lon_lat_to_cartesian_ecef(cell_corner_lon[0], cell_corner_lat[0])
        xyz2 = self.lon_lat_to_cartesian_ecef(cell_corner_lon[1], cell_corner_lat[1])
        
        # Reshape
        xyz1, xyz2 = xyz1[0], xyz2[0]

        for i in range(2, spatial_nv):

            xyz3 = self.lon_lat_to_cartesian_ecef(cell_corner_lon[i], cell_corner_lat[i])

            # Reshape
            xyz3 = xyz3[0]

            sum += self.mod_tri_area(xyz1, xyz2, xyz3)
            
            # Copy
            xyz2 = deepcopy(xyz3)

    return sum

def mod_tri_area(self, u, v, w):
    
    tmp_vec = self.cross_product(u, v)
    sina = np.sqrt(self.norm(tmp_vec))
    a = np.arcsin(sina)

    tmp_vec = self.cross_product(u, w)
    sinb = np.sqrt(self.norm(tmp_vec))
    b = np.arcsin(sinb)

    tmp_vec = self.cross_product(w, v)
    sinc = np.sqrt(self.norm(tmp_vec))
    c = np.arcsin(sinc)

    s = 0.5*(a+b+c)

    t = np.tan(s*0.5) * np.tan((s - a)*0.5) * np.tan((s - b)*0.5) * np.tan((s - c)*0.5)

    area = np.fabs(4.0 * np.arctan(np.sqrt(np.fabs(t))))

    return area

@staticmethod
def cross_product(a, b):
    
    return [a[1]*b[2] - a[2]*b[1], a[2]*b[0] - a[0]*b[2], a[0]*b[1] - a[1]*b[0]]

@staticmethod
def norm(a):

    return (a[0]*a[0] + a[1]*a[1] + a[2]*a[2])

